'''
Author: Alexandre Miguel de Carvalho <carvalho@machinelearning.com>
Author: Ronaldo Cristiano Prati <ronaldo.prati@ufabc.edu.br>

MIT License
Copyright (c) 2020 Alexandre Miguel de Carvalho and Ronaldo Cristiano Prati

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from sklearn.decomposition import PCA
import os
from collections import Counter
import numpy as np
import pandas as pd
from imblearn.over_sampling.base import BaseOverSampler
from scipy.spatial import Delaunay
from sklearn.utils import check_random_state
from DTOSMOTE.TetrahedronStatistics import tetrahedron_stats


class DTO(BaseOverSampler):
	
	def __init__(self,
	             dataset_name,
	             geometry,
	             dirichlet,
	             sampling_strategy='auto',
	             alpha=1,
	             n_jobs=-1,
	             diff_alpha=0.5,
	             random_state=None,
	             pca_s=3):
		'''
		
		:param dataset_name: dataset name
		:param geometry: 'area', 'volume','area_volume_ratio','edge_ratio','radius_ratio','aspect_ratio',
		  'max_solid_angle','min_solid_angle','solid_angle'
		:param dirichlet: float[1,10]
		:param pca_s: PCA components, default = 3
		'''
		
		super(DTO, self).__init__(sampling_strategy=sampling_strategy)
		
		self.dim_redutor = PCA()
		self.alpha = alpha
		self.n_jobs = n_jobs
		self.mesh = None
		self.order = geometry
		self.diff_alpha = diff_alpha
		self.equal_alpha = dirichlet
		self.dataset_name = dataset_name
		self.random_state = random_state
		self.pca_s = pca_s
		self.mesh_folder = dataset_name
		self._make_dir()
	
	def _fit_resample(self, X, y):
		return self._sample(X, y)
	
	def _make_dir(self):
		try:
			os.makedirs(self.dataset_name)
		except:
			pass
		self.mash_folder = self.dataset_name
	
	def build_mesh_3D(self, X, y):
		self.dim_redutor.set_params(n_components=self.pca_s)
		X_reduced = self.dim_redutor.fit_transform(X)
		self.mesh = Delaunay(X_reduced)
		np.savetxt('./'+self.mesh_folder+'/' + self.dataset_name + '_simplices.csv', self.mesh.simplices)
		np.savetxt('./'+self.mesh_folder+'/' + self.dataset_name + '_points.csv', self.mesh.points)
		self.tetra_props = pd.DataFrame([Counter(y[simplex])
		                                 for simplex in self.mesh.simplices]).fillna(0) / 4
		self.tetra_props.to_csv('./'+self.mesh_folder+'/' + self.dataset_name + '_tetra_props.csv', index=False)
		self.tetra_stats = pd.DataFrame([tetrahedron_stats(X_reduced[simplex]).values()
		                                 for simplex in self.mesh.simplices]).fillna(0)
		self.tetra_stats.to_csv('./'+self.mesh_folder+'/' + self.dataset_name + '_tetra_stats.csv', index=False)
	
	def weight_tetrahedrons(self, class_sample, order):
		if order == "random":
			w = self.tetra_props[class_sample]
		else:
			w = self.tetra_props[class_sample] * self.tetra_stats[order]  # class_sample==1 (minority)
		
		return (w / sum(w))
	
	def beta(self, y_values, class_sample):
		beta = np.ones(len(y_values)) * self.diff_alpha;
		beta[y_values == class_sample] = self.equal_alpha;
		return self.random_state.dirichlet(beta, 1)[0]
	
	def genPoint(self, x_frame, beta):
		'''
		Feature generation oversampling
		:param x_frame: matrix with original X data vector
		:param beta: for numbers in a vector with total sum = 1 for example beta = [0.25, 0.25, 0.25, 0.25]
		:return: new feature vector
		'''
		return np.dot(beta.T, x_frame)
	
	def _sample(self, X, y):
		
		if self.mesh == None:
			self.build_mesh_3D(X, y)  # mesh and properties
		X_resampled = X.copy()
		y_resampled = y.copy()
		self.random_state = check_random_state(self.random_state)
		# class_sample is the label of each class
		# n_samples is the necessary samples to balanced this class
		for class_sample, n_samples in self.sampling_strategy_.items():
			if n_samples == 0:
				continue
			
			probs = self.weight_tetrahedrons(class_sample, self.order)
			indexnotz = np.nonzero(probs.values)
			simplices_to_interpolate = self.random_state.choice(indexnotz[0], n_samples,
			                                                    p=np.take(probs, indexnotz[0], axis=0))
			X_new = np.array(
					[self.genPoint(X[self.mesh.simplices[i]], self.beta(y[self.mesh.simplices[i]], class_sample))
					 for i in simplices_to_interpolate])
			y_new = np.array([class_sample] * n_samples)
			X_resampled = np.vstack((X_resampled, X_new))
			y_resampled = np.hstack((y_resampled, y_new))
		
		return (X_resampled, y_resampled)



'''
Author: Alexandre Miguel de Carvalho <carvalho@machinelearning.com>
Author: Ronaldo Cristiano Prati <ronaldo.prati@ufabc.edu.br>

MIT License
Copyright (c) 2020 Alexandre Miguel de Carvalho and Ronaldo Cristiano Prati

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

from itertools import combinations
import numpy as np
from scipy.spatial import ConvexHull
from scipy.spatial.distance import pdist


class tetrahedron_stats:
	
	def __init__(self, pts):
		self.labels = "ABCD"
		self.hull = ConvexHull(pts)
		self.pts = dict(zip(self.labels, pts))
		self.dists = pdist(pts)
		self.angles = self.tetrahedron_solid_angles()
	
	def area(self):
		return self.hull.area
	
	def volume(self):
		return self.hull.volume
	
	def edge_ratio(self):
		return max(self.dists) / min(self.dists)
	
	def radius_ratio(self):
		circumcenter = self.calc_circumcenter()
		innercenter = self.calc_incencenter()
		return (3 * innercenter['radius'] / circumcenter['radius'])
	
	def aspect_ratio(self):
		E = max(self.dists)
		R = self.calc_circumcenter()['radius']
		return (R / E)
	
	def tetrahedron_edges(self):
		return ({"".join([i, j]): self.pts[i] - self.pts[j] for i, j in combinations(self.labels, 2)})
	
	def normals(self, edges):
		pairs = [("".join((l[0], l[1])), "".join((l[0], l[2]))) for l in combinations(self.labels, 3)]
		normals = [np.cross(edges[e1], edges[e2]) for (e1, e2) in pairs]
		return (dict(zip(self.labels, normals)))
	
	def angle(self, u, v):
		dot = np.dot(u, v)
		cos_theta = dot / (np.linalg.norm(u) * np.linalg.norm(v))
		return (np.arccos(cos_theta))
	
	def tetrahedron_dihedral_angles(self):
		edges = self.tetrahedron_edges()
		normals = self.normals(edges)
		angles = {"".join([i, j]): np.pi - self.angle(normals[i], normals[j]) for i, j in combinations(self.labels, 2)}
		return angles
	
	def tetrahedron_solid_angles(self):
		dihedral_angles = self.tetrahedron_dihedral_angles()
		triangles = [["".join(edge) for edge in combinations(facet, 2)] for facet in combinations(self.labels, 3)]
		return [np.sum([dihedral_angles[edge] for edge in triangle]) - np.pi for triangle in triangles]
	
	def max_solid_angle(self):
		return (max(self.angles))
	
	def min_solid_angle(self):
		return (min(self.angles))
	
	def solid_angle(self):
		return (sum(self.angles))
	
	def calc_circumcenter(self):
		""" Calculates the cirumcenters of the circumspheres of tetrahedrons.
		An implementation based on
		http://mathworld.wolfram.com/Circumsphere.html
		"""
		tetrahedron = np.array(list(self.pts.values()))
		a = np.concatenate((tetrahedron, np.ones((4, 1))), axis=1)
		sums = np.sum(tetrahedron ** 2, axis=1)
		d = np.concatenate((sums[:, np.newaxis], a), axis=1)
		c = np.concatenate((sums[:, np.newaxis], tetrahedron), axis=1)
		dx = np.delete(d, 1, axis=1)
		dy = np.delete(d, 2, axis=1)
		dz = np.delete(d, 3, axis=1)
		dx = np.linalg.det(dx)
		dy = -np.linalg.det(dy)
		dz = np.linalg.det(dz)
		a = np.linalg.det(a)
		c = np.linalg.det(c)
		nominator = np.array((dx, dy, dz))
		denominator = 2 * a
		radius = np.sqrt(sum(nominator ** 2) - 4 * a * c) / np.abs(denominator)
		
		return {"center": (nominator / denominator).T, "radius": radius}
	
	def calc_incencenter(self):
		tetrahedron = np.array(list(self.pts.values()))
		edges = self.tetrahedron_edges()
		normals = self.normals(edges)
		n = np.array(list(normals.values()))
		sums = np.sum(n ** 2, axis=1)
		A = np.concatenate((n, sums[:, np.newaxis]), axis=1)
		u = np.sum(np.multiply(n, tetrahedron), axis=1)
		s = np.linalg.solve(A, u)
		
		return {"center": s[0:3], "radius": np.abs(s[3])}
	
	def values(self):
		return ({'area': self.area(),
				 'volume': self.volume(),
				 'area_volume_ratio': self.volume() / self.area(),
				 'edge_ratio': self.edge_ratio(),
				 'radius_ratio': self.radius_ratio(),
				 'aspect_ratio': self.aspect_ratio(),
				 'max_solid_angle': self.max_solid_angle(),
				 'min_solid_angle': self.min_solid_angle(),
				 'solid_angle': self.solid_angle()
				 })



from collections import Counter
from sklearn.datasets import make_classification
from DTOSMOTE.dto_smote import DTO

def test_dto():
    X, y = make_classification(n_classes=3, class_sep=2,
                           weights=[0.1,0.2, 0.9], n_informative=3, n_redundant=1, flip_y=0,
                           n_features=20, n_clusters_per_class=1, n_samples=100, random_state=10)
    
    print('Original dataset shape: ',Counter(y))
    delaunay = DTO('teste1','solid_angle',7.5)
    X_res, y_res = delaunay.fit_resample(X, y)
    print('Resampled dataset shape: ',Counter(y_res))



if __name__ == "__main__":
    test_dto()


from distutils.core import setup

setup(
		name='DTOSMOTE',
		packages=['DTOSMOTE'],
		version='1.0.0',  # Ideally should be same as your GitHub release tag varsion
		description='Delaunay Thetraedral Oversampling SMOTE',
		author='Alexandre Miguel de Carvalho & Ronaldo Cristiano Prati',
		author_email='',
		url='https://github.com/carvalhoamc/DTOSMOTE',
		download_url='https://github.com/carvalhoamc/DTOSMOTE/archive/v1.0.tar.gz',
		keywords=['DTOSMOTE', 'oversampling'],
		classifiers=[  # Optional
			# How mature is this project? Common values are
			#   3 - Alpha
			#   4 - Beta
			#   5 - Production/Stable
			'Development Status :: 5 - Production/Stable',
			# Pick your license as you wish
			'License :: OSI Approved :: MIT License',
			
			# Specify the Python versions you support here. In particular, ensure
			# that you indicate whether you support Python 2, Python 3 or both.
			'Programming Language :: Python :: 3',
			'Programming Language :: Python :: 3.4',
			'Programming Language :: Python :: 3.5',
			'Programming Language :: Python :: 3.6',
		],

)
